import sys
from typing import Any

from aqueduct_executor.operators.connectors.data import common, config, connector
from aqueduct_executor.operators.connectors.data.execute import execute_data_spec
from aqueduct_executor.operators.connectors.data.spec import Spec
from aqueduct_executor.operators.spark.utils import read_artifacts_spark, write_artifact_spark
from aqueduct_executor.operators.utils.exceptions import (
    MissingConnectorDependencyException,
    UnsupportedConnectorExecption,
)
from pyspark.sql import SparkSession


def run(spec: Spec, spark_session_obj: SparkSession) -> None:
    """
    Runs one of the following connector operations:
    - authenticate
    - extract
    - load
    - load-table
    - delete-saved-objects
    - discover

    Arguments:
    - spec: The spec provided for this operator.
    - spark_session_obj: The SparkSession
    """
    return execute_data_spec(
        spec=spec,
        read_artifacts_func=read_artifacts_spark,
        write_artifact_func=write_artifact_spark,
        setup_connector_func=setup_connector_spark,
        is_spark=True,
        spark_session_obj=spark_session_obj,
    )


def setup_connector_spark(
    connector_name: common.Name, connector_config: config.Config
) -> connector.DataConnector:
    """
    Sets up the Spark Connectors. We currently support the following resources with Spark:
    - S3
    - Snowflake
    """
    # prevent isort from moving around type: ignore comments which will cause mypy issues.
    # isort: off
    if connector_name == common.Name.SNOWFLAKE:
        try:
            from pyspark.sql import SparkSession
            from snowflake import sqlalchemy
        except:
            raise MissingConnectorDependencyException(
                "Unable to initialize the Spark Snowflake connector. Have you run `aqueduct install spark-snowflake`?"
            )

        from aqueduct_executor.operators.connectors.data.spark.snowflake import (
            SparkSnowflakeConnector as OpConnector,
        )
    elif connector_name == common.Name.S3:
        try:
            import pyarrow
        except:
            raise MissingConnectorDependencyException(
                "Unable to initialize the Spark S3 connector. Have you run `aqueduct install s3`?"
            )

        from aqueduct_executor.operators.connectors.data.spark.s3 import (  # type: ignore
            SparkS3Connector as OpConnector,
        )
    else:
        raise UnsupportedConnectorExecption(
            "Unable to initialize connector. This connector is not yet supported for Aqueduct on Spark."
        )
    # isort: on
    return OpConnector(config=connector_config)  # type: ignore



import functools
import io
import json
from typing import Any, Dict, List, Optional

import cloudpickle as pickle
import numpy as np
import pandas as pd
from aqueduct_executor.operators.connectors.data import common, connector, extract, load, s3
from aqueduct_executor.operators.connectors.data.config import S3Config
from aqueduct_executor.operators.connectors.data.utils import construct_boto_session
from aqueduct_executor.operators.utils.enums import ArtifactType
from aqueduct_executor.operators.utils.saved_object_delete import SavedObjectDelete
from aqueduct_executor.operators.utils.utils import delete_object
from botocore.client import ClientError
from PIL import Image
from pyspark.sql import DataFrame, SparkSession

_DEFAULT_JSON_ENCODING = "utf8"
_DEFAULT_IMAGE_FORMAT = "jpeg"
s3_template = "s3a://%s/%s"


class SparkS3Connector(s3.S3Connector):
    def __init__(self, config: S3Config):
        super().__init__(config)

    def _fetch_object_spark(
        self, key: str, params: extract.S3Params, spark_session_obj: SparkSession
    ) -> Any:
        # Table artifacts use spark to load data into Spark DataFrames.
        if params.artifact_type == ArtifactType.TABLE:
            if params.format is None:
                raise Exception("You must specify a file format for table data.")
            data_path = s3_template % (self.bucket, key)
            try:
                if params.format == common.S3TableFormat.CSV:
                    return spark_session_obj.read.option("header", "true").csv(data_path)
                elif params.format == common.S3TableFormat.JSON:
                    return spark_session_obj.read.json(data_path)
                elif params.format == common.S3TableFormat.PARQUET:
                    return spark_session_obj.read.parquet(data_path)
            except Exception:
                raise Exception(
                    "Unable to read in table at path `%s` with S3 file format `%s`."
                    % (key, params.format)
                )
            else:
                raise Exception(
                    "Unknown S3 file format `%s` for file at path `%s`." % (params.format, key)
                )
        # Non-table artifacts use same serialization as regular S3 integration.
        else:
            return self.fetch_object(key, params)

    def extract_spark(self, params: extract.S3Params, spark_session_obj: SparkSession) -> Any:
        path = json.loads(params.filepath)
        if not isinstance(path, List):
            if len(path) == 0:
                raise Exception("S3 file path cannot be an empty string.")
            if path[-1] == "/":
                # This means the path is a directory, and we will do a prefix search.
                files = []
                for obj in self.s3.Bucket(self.bucket).objects.filter(Prefix=path):
                    # The filter api also returns the directories, so we filter them out.
                    if (obj.key)[-1] != "/":
                        files.append(self._fetch_object_spark(obj.key, params, spark_session_obj))

                if params.artifact_type == ArtifactType.TABLE and params.merge:
                    # We ignore indexes anyways when serializing the data later, so it's ok to do it earlier here.
                    return pd.concat(files, ignore_index=True)
                else:
                    return tuple(files)
            else:
                # This means the path is a file name, and we do a regular file retrieval.
                return self._fetch_object_spark(path, params, spark_session_obj)
        else:
            # This means we have a list of file paths.
            files = []
            for key in path:
                if len(key) == 0:
                    raise Exception("S3 file path cannot be an empty string.")
                if key[-1] == "/":
                    raise Exception("Each key in the list must not be a directory, found %s." % key)
                files.append(self._fetch_object_spark(key, params, spark_session_obj))

            if params.artifact_type == ArtifactType.TABLE and params.merge:
                # We ignore indexes anyways when serializing the data later, so it's ok to do it earlier here.
                return self.unionAll(files)
            else:
                return tuple(files)

    def load_spark(self, params: load.S3Params, data: Any, artifact_type: ArtifactType) -> None:
        if artifact_type == ArtifactType.TABLE:
            if params.format is None:
                raise Exception("You must specify a file format for table data.")

            # data is a Spark DataFrame.
            data_path = s3_template % (self.bucket, params.filepath)
            if params.format == common.S3TableFormat.CSV:
                data.write.csv(data_path)
            elif params.format == common.S3TableFormat.JSON:
                data.write.json(data_path)
            elif params.format == common.S3TableFormat.PARQUET:
                data.write.parquet(data_path)
            else:
                raise Exception("Unknown S3 file format %s." % params.format)
        else:
            # data is not a Spark DataFrame, use normal S3 integration's load.
            self.load(params, data, artifact_type)

    def unionAll(self, dfs: Any) -> Any:
        return functools.reduce(lambda df1, df2: df1.union(df2.select(df1.columns)), dfs)



import io
import json
from typing import Any, Callable, Dict, List, Optional

import cloudpickle as pickle
import pandas as pd
from aqueduct.constants.enums import S3SerializationType
from aqueduct.utils.serialization import PickleableCollectionSerializationFormat
from aqueduct.utils.type_inference import infer_artifact_type
from aqueduct_executor.operators.connectors.data.common import S3TableFormat
from aqueduct_executor.operators.utils.enums import ArtifactType
from PIL import Image

_DEFAULT_JSON_ENCODING = "utf8"
_DEFAULT_IMAGE_FORMAT = "jpeg"


def _read_csv_table(content: bytes) -> Any:
    return pd.read_csv(io.BytesIO(content))


def _read_json_table(content: bytes) -> Any:
    return pd.read_json(io.BytesIO(content))


def _read_parquet_table(content: bytes) -> Any:
    return pd.read_parquet(io.BytesIO(content))


def _read_json_content(content: bytes) -> Any:
    # This assumes that the encoding is "utf-8". May worth considering letting the user
    # specify custom encoding in the future.
    json_data = content.decode(_DEFAULT_JSON_ENCODING)
    # Make sure the data is a valid json object.
    json.loads(json_data)
    return json_data


def _read_bytes_content(content: bytes) -> Any:
    return content


def _read_image_content(content: bytes) -> Any:
    return Image.open(io.BytesIO(content))


def _read_pickle_content(content: bytes) -> Any:
    return pickle.loads(content)


_s3_deserialization_function_mapping: Dict[str, Callable[[bytes], Any]] = {
    S3SerializationType.CSV_TABLE: _read_csv_table,
    S3SerializationType.JSON_TABLE: _read_json_table,
    S3SerializationType.PARQUET_TABLE: _read_parquet_table,
    S3SerializationType.JSON: _read_json_content,
    S3SerializationType.BYTES: _read_bytes_content,
    S3SerializationType.IMAGE: _read_image_content,
    S3SerializationType.PICKLE: _read_pickle_content,
}


def _write_csv_table(output: pd.DataFrame) -> bytes:
    buf = io.BytesIO()
    output.to_csv(buf, index=False)
    return buf.getvalue()


def _write_json_table(output: pd.DataFrame) -> bytes:
    buf = io.BytesIO()
    # Index cannot be False for `to.json` for default orient
    # See: https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.to_json.html
    output.to_json(buf)
    return buf.getvalue()


def _write_parquet_table(output: pd.DataFrame) -> bytes:
    buf = io.BytesIO()
    output.to_parquet(buf, index=False)
    return buf.getvalue()


def _write_json_content(output: str) -> bytes:
    return output.encode(_DEFAULT_JSON_ENCODING)


def _write_bytes_content(output: bytes) -> bytes:
    return output


def _write_image_content(output: Image.Image) -> bytes:
    img_bytes = io.BytesIO()
    output.save(img_bytes, format=_DEFAULT_IMAGE_FORMAT)
    return img_bytes.getvalue()


def _write_pickle_content(output: Any) -> bytes:
    return bytes(pickle.dumps(output))


__s3_serialization_function_mapping: Dict[str, Callable[..., bytes]] = {
    S3SerializationType.CSV_TABLE: _write_csv_table,
    S3SerializationType.JSON_TABLE: _write_json_table,
    S3SerializationType.PARQUET_TABLE: _write_parquet_table,
    S3SerializationType.JSON: _write_json_content,
    S3SerializationType.BYTES: _write_bytes_content,
    S3SerializationType.IMAGE: _write_image_content,
    S3SerializationType.PICKLE: _write_pickle_content,
}


def serialize_val_for_s3(
    val: Any,
    serialization_type: S3SerializationType,
    format: Optional[S3TableFormat],
) -> bytes:
    """Serializes a computed value into bytes to be uploaded to S3.

    Unlike deserialization, the serialization of data to S3 relies on potentially
    inferring the artifact and serialization type of elements in a list/tuple. Since
    the artifact -> serialization mapping is different for S3, it's cleaner to just
    implement an S3 version of serialization.
    """
    if serialization_type == S3SerializationType.PICKLE and (
        isinstance(val, list) or isinstance(val, tuple)
    ):
        elem_serialization_types: List[S3SerializationType] = []
        for elem in val:
            elem_artifact_type = infer_artifact_type(elem)

            elem_serialization_types.append(
                artifact_type_to_s3_serialization_type(
                    elem_artifact_type,
                    format,
                    ignore_format_requirement=True,
                )
            )

        data: List[bytes] = [
            __s3_serialization_function_mapping[elem_serialization_types[i]](val[i])
            for i in range(len(elem_serialization_types))
        ]

        pickled_collection_data = PickleableCollectionSerializationFormat(
            aqueduct_serialization_types=elem_serialization_types,
            data=data,
            is_tuple=isinstance(val, tuple),
        )

        # The value we end up pickling is a dictionary.
        val = pickled_collection_data.dict()

    return __s3_serialization_function_mapping[serialization_type](val)


class S3UnsupportedArtifactTypeException(Exception):
    pass


class S3UnknownFileFormatException(Exception):
    pass


def artifact_type_to_s3_serialization_type(
    artifact_type: ArtifactType,
    format: Optional[S3TableFormat],
    ignore_format_requirement: bool = False,
) -> S3SerializationType:
    if artifact_type == ArtifactType.TABLE:
        if format is None:
            if not ignore_format_requirement:
                raise Exception("You must specify a file format for table data.")

            # Default format is Parquet.
            format = S3TableFormat.PARQUET

        if format == S3TableFormat.CSV:
            return S3SerializationType.CSV_TABLE
        elif format == S3TableFormat.JSON:
            return S3SerializationType.JSON_TABLE
        elif format == S3TableFormat.PARQUET:
            return S3SerializationType.PARQUET_TABLE
        else:
            raise S3UnknownFileFormatException("Unknown S3 file format `%s`" % format)
    elif artifact_type == ArtifactType.JSON:
        serialization_type = S3SerializationType.JSON
    elif artifact_type == ArtifactType.IMAGE:
        serialization_type = S3SerializationType.IMAGE
    elif artifact_type == ArtifactType.BYTES:
        serialization_type = S3SerializationType.BYTES

    elif (
        artifact_type == ArtifactType.STRING
        or artifact_type == ArtifactType.BOOL
        or artifact_type == ArtifactType.NUMERIC
        or artifact_type == ArtifactType.DICT
        or artifact_type == ArtifactType.LIST
        or artifact_type == ArtifactType.TUPLE
        or artifact_type == ArtifactType.PICKLABLE
    ):
        serialization_type = S3SerializationType.PICKLE
    else:
        raise S3UnsupportedArtifactTypeException("Unsupported data type %s." % artifact_type)

    assert serialization_type is not None, (
        "Unimplemented case for artifact type `%s`" % artifact_type
    )
    return serialization_type



from typing import Dict

import pandas as pd
from aqueduct_executor.operators.connectors.data import config, relational, utils
from sqlalchemy import create_engine, engine
from sqlalchemy.types import VARCHAR


class SnowflakeConnector(relational.RelationalConnector):
    def __init__(self, config: config.SnowflakeConfig):
        conn_engine = create_snowflake_engine(config)
        super().__init__(
            conn_engine=conn_engine,
            object_to_varchar_mapper=map_object_dtype_to_varchar,
        )


def create_snowflake_engine(config: config.SnowflakeConfig) -> engine.Engine:
    # Snowflake Dialect:
    # https://github.com/snowflakedb/snowflake-sqlalchemy
    url = "snowflake://{username}:{password}@{account_identifier}/{database}/{schema}?warehouse={warehouse}".format(
        username=config.username,
        password=utils.url_encode(config.password),
        account_identifier=config.account_identifier,
        database=config.database,
        schema=config.db_schema,
        warehouse=config.warehouse,
    )

    if config.role:
        # Assume a role explicitly, because one has been provided
        url += "&role={role}"
        url = url.format(role=config.role)

    return create_engine(url)


def map_object_dtype_to_varchar(df: pd.DataFrame) -> Dict[str, VARCHAR]:
    col_to_type = {}
    for col in df.select_dtypes(include=["object"]):
        # We do not need to provide an explicit size, because Snowflake
        # will default to allowing the maximum number of characters.
        # There is no performance difference between using the full-length VARCHAR
        # declaration VARCHAR(16777216) and a smaller length.
        # See: https://docs.snowflake.com/en/sql-reference/data-types-text
        col_to_type[col] = VARCHAR()
    return col_to_type
